/*

	ICONBUTT.C

	Gestor de botones tipo Icono.

	(C) Angel Ortega 1991/1994 Kaplan Development Software.


	NOTAS:
	- Asume windows.h includo.
	- Se debe incluir la funcin IBReg() en la primera instancia
	  del programa.
	- En el fichero de recursos, el caption del icono debe ser
	  un nombre de icono en el mismo ejecutable.
	- El nombre de la nueva clase es "IconButton".

*/


/* tamao en pxels de botn tipo IconButton */
#define IB_SIZE 36


void Rect3D(HDC hdc, RECT * rect, int tipo)
/* dibuja en el hdc un rectngulo 3D, segn los tipos:
	0 - Liberado (convexo).
	1 - Pulsado (cncavo).
	2 - Inhabilitado (plano). */
{
	HPEN hp;
	HANDLE hop;

	/* se crea una pluma gris oscuro */
	hp=CreatePen(PS_SOLID, 1, 0x007F7F7F);
	hop=SelectObject(hdc, hp);

	switch(tipo)
	{
	case 0:
	/* liberado */

		/* se dibuja el borde oscuro */
		MoveTo(hdc, rect->left+1, rect->bottom-2);
		LineTo(hdc, rect->left+2, rect->bottom-3);
		LineTo(hdc, rect->left+1, rect->bottom-2);
		LineTo(hdc, rect->right-2, rect->bottom-2);
		LineTo(hdc, rect->right-3, rect->bottom-3);
		LineTo(hdc, rect->right-2, rect->bottom-2);
		LineTo(hdc, rect->right-2, rect->top+2);

		/* se selecciona un color blanco */
		SelectObject(hdc, GetStockObject(WHITE_PEN));

		/* se borra la pluma */
		DeleteObject(hp);

		/* se dibuja un borde claro */
		MoveTo(hdc, rect->right-2, rect->top+1);
		LineTo(hdc, rect->right-3, rect->top+2);
		LineTo(hdc, rect->right-2, rect->top+1);
		LineTo(hdc, rect->left+1, rect->top+1);
		LineTo(hdc, rect->left+2, rect->top+2);
		LineTo(hdc, rect->left+1, rect->top+1);
		LineTo(hdc, rect->left+1, rect->bottom-2);

		/* se recupera la pluma inicial */
		SelectObject(hdc, hop);

		break;

	case 1:
	/* pulsado */

		MoveTo(hdc, rect->left+1, rect->top+1);
		LineTo(hdc, rect->left+1, rect->bottom-1);
		MoveTo(hdc, rect->left+1, rect->top+1);
		LineTo(hdc, rect->right-1, rect->top+1);

		MoveTo(hdc, rect->left+2, rect->top+2);
		LineTo(hdc, rect->left+2, rect->bottom-1);
		MoveTo(hdc, rect->left+2, rect->top+2);
		LineTo(hdc, rect->right-1, rect->top+2);

		/* se borra la pluma */
		SelectObject(hdc, hop);
		DeleteObject(hp);

		break;

	case 2:
	/* plano */

		MoveTo(hdc, rect->left+3, rect->bottom-3);
		LineTo(hdc, rect->left+3, rect->top+3);
		LineTo(hdc, rect->right-3, rect->top+3);

		/* se selecciona un color blanco */
		SelectObject(hdc, GetStockObject(WHITE_PEN));

		/* se borra la pluma */
		DeleteObject(hp);

		/* se dibuja un borde claro */
		MoveTo(hdc, rect->left+3, rect->bottom-3);
		LineTo(hdc, rect->right-3, rect->bottom-3);
		LineTo(hdc, rect->right-3, rect->top+3);

		/* se recupera la pluma inicial */
		SelectObject(hdc, hop);
	}
}


long FAR PASCAL IBProc(HWND hwnd, WORD msg, WORD wParam, LONG lParam)
/* gestor de los botones de icono */
{
	char strtmp[100];
	HICON h;
	HDC hdc;
	PAINTSTRUCT ps;
	RECT rect;
	HANDLE hob;
	HBRUSH hb;
	int x;

	int estado=0;

	switch(msg)
	{
	case WM_CREATE:

		/* modo no pulsado */
		SendMessage(hwnd, BM_SETSTATE, 0, 0L);
		break;

	case WM_PAINT:

		if(SendMessage(hwnd, BM_GETSTATE, 0, 0L))
			estado=1;
		else
			estado=0;

		/* se lee el caption: ste ser el nombre del icono */
		GetWindowText(hwnd, strtmp, sizeof(strtmp));

		/* se carga (hInst debe ser global) */
		h=LoadIcon(hInst, strtmp);

		/* se comienza a pintar */
		hdc=BeginPaint(hwnd, &ps);

		/* se crea una brocha gris claro */
		hb=CreateSolidBrush(0x00C0C0C0);

		hob=SelectObject(hdc, hb);

		/* se dibuja el borde */
		GetClientRect(hwnd, &rect);
		Rectangle(hdc, 0, 0, rect.right-rect.left,
			rect.bottom-rect.top);

		/* se selecciona el viejo */
		SelectObject(hdc, hob);

		/* se borra la brocha gris claro ... */
		DeleteObject(hb);

		if(!IsWindowEnabled(hwnd))
			estado=2;

		/* se dibuja el rectngulo 3D */
		Rect3D(hdc, &rect, estado);
		estado&=1;

		x=(rect.right-rect.left)/2;
		x-=GetSystemMetrics(SM_CXICON)/2;

//		if(x<0)
//			x=0;

		/* se dibuja el icono */
		DrawIcon(hdc, x+estado*2, x+estado*2, h);

		EndPaint(hwnd, &ps);

		return(0);

	case WM_SETTEXT:

		/* si se cambia el nombre (el icono), se invalida el rea
		   de dibujo */
		InvalidateRect(hwnd, NULL, FALSE);
		break;

	case WM_ENABLE:

		/* tambin se invalida el aspecto del icono */
		InvalidateRect(hwnd, NULL, FALSE);
		break;

	case WM_KEYUP:

		/* pulsacin de tecla: espacio? */
		if(wParam != VK_SPACE)
			break;

		/* si es espacio contina ... */
	case WM_LBUTTONUP:

		ReleaseCapture();

		/* botn no activado */
		SendMessage(hwnd, BM_SETSTATE, 0, 0L);

		/* se invalida, para que se redibuje */
		InvalidateRect(hwnd, NULL, FALSE);

		/* se informa de la pulsacin al padre */
		SendMessage(GetParent(hwnd), WM_COMMAND,
			GetWindowWord(hwnd, GWW_ID), (LONG) hwnd);

		return(0);

	case WM_KEYDOWN:
	case WM_LBUTTONDOWN:

		/* se captura el ratn */
		SetCapture(hwnd);

		/* botn activado */
		SendMessage(hwnd, BM_SETSTATE, 1, 0L);

		/* se invalida, para que se redibuje */
		InvalidateRect(hwnd, NULL, FALSE);

		return(0);

	case BM_SETSTATE:

		SetWindowWord(hwnd, 0, wParam);
		return(0);;

	case BM_GETSTATE:
		return(GetWindowWord(hwnd, 0));
	}

	return DefWindowProc(hwnd, msg, wParam, lParam);
}


void IBReg(HANDLE hInstance)
/* registra el tipo de botn de icono */
{
	WNDCLASS wc;

	wc.style=CS_HREDRAW|CS_VREDRAW;
	wc.lpfnWndProc=(WNDPROC) IBProc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=sizeof(int);
	wc.hInstance=hInstance;
	wc.hIcon=NULL;
	wc.hCursor=LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground=COLOR_BTNFACE+1;
	wc.lpszMenuName=NULL;
	wc.lpszClassName="IconButton";

	RegisterClass(&wc);
}
