/*

        WAIMOTOR.C

        Nucleo Central de The Wailer.
        Version 2.0.

        (C) 1993/1995 Angel Ortega.


        Variables:

        int Conferencia;
        long Mensaje;
        char Conferencias[MAX_CONFS][MAX_NOM_CONF];
        int NumConfes;
        int NumMsgs;
        int Histo;
        char To[150];
        char Remit[150];
        char Subject[100];
        char Fecha[150];
        int ancho_linea=75;
        char Yo[30]="";
        int use_histo;
        int ObjType;


        Funciones:

        void SetMsgFlags(unsigned char modo);

        int ReadMsg(char * hptr, int hsize, char * bptr, int bsize,
                void (* procstr)(char * cadena));

        int QuoteMsg(char * bptr, int bsize,
                void (* procstr)(char * cadena));

        int WriteMsg(char * bptr, char * conf);

        int ReadList(void (* newelem)(void));

        int Index(char * dat, char * ind);

        void ProcessNew(void);

        void InitRespFile(void);

        int InitW(void);

        void EndW(void);

*/

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>


#if !defined(O_BINARY)
        #define O_BINARY 0
        #define READ_BINARY "r"
        #define NL_STR "\n"
#else
        #define READ_BINARY "rb"
        #include <io.h>
        #define NL_STR "\r\n"
#endif


#if !defined(min)
        #define min(x,y) ((x)<(y))?(x):(y)
#endif


/*******************
     Constantes
*******************/

/* ningun mensaje seleccionado */
#define NO_MENSAJE 0xFFFFFFFF

/* caracter(es) de quote */
#define QUOTE_CHARS "> "

/* fichero de conferencias */
#define CONFE_FILE "conferen.dat"
/* mensajes nuevos */
#define NUEVOS_DAT "nuevos.dat"
/* historico de mensajes */
#define HISTO_DAT "historic.dat"
/* indice de nuevos */
#define NUEVOS_IND "nuevos.ind"
/* indice del historico */
#define HISTO_IND "historic.ind"
/* fichero temporal */
#define TMP_FILE "wailer.tmp"
/* fichero con la firma */
#define FIRM_FILE "wailer.fir"
/* fichero con los tags */
#define TAG_FILE "wailer.tag"

/* mascara de borrado */
#define MASK_BOR 0x80

/* mascara de leido */
#define MASK_LEI 0x40

/* mascara de conferencia */
#define MASK_CONF 0x3F

/* maximo de conferencias */
#define MAX_CONFS 64

/* maximo del nombre de conferencia */
#define MAX_NOM_CONF 40

/* MIME */

/* boundary de los mensajes multipart/mixed MIME */
char * WailerBoundary="___boundary___";

/* tabla de caracteres en base64 (MIME) */
char Base64[66]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"\
		"0123456789+/=";

/* tokens MIME */
char * MimeContentType="X-Content-Type: ";
char * MimeConTraEnc="X-Content-Transfer-Encoding: ";
char * MimeConDes="X-Content-Description: ";


/*************************
         Globales
*************************/

/* nombre del programa (general) */
char * APPGEN="The Wailer";

/* conferencia seleccionada */
int Conferencia=-1;

/* mensaje seleccionado */
long Mensaje=NO_MENSAJE;

/* array con las conferencias */
char Conferencias[MAX_CONFS][MAX_NOM_CONF];

/* numero de conferencias */
int NumConfes=0;

/* numero de mensajes en esta conferencia */
int NumMsgs=0;

/* flag de historico: 1, mensajes viejos; 0, nuevos */
int Histo=0;

/* estructura del indice de mensajes */
struct tagind
{
        unsigned char dat;      /* bits 0-5: numero de conferencia;
                                   bit 6: leido;
                                   bit 7: borrado */
        long offset;            /* offset del principio del mensaje
                                   en el disco */
        long tamcab;    	/* tamano de la cabecera del mensaje */
	long tambody;   	/* tamano del cuerpo del mensaje */
        unsigned int offsub;    /* offset del subject */
        unsigned int offfrom;   /* offset del remitente */
        unsigned int offto;     /* offset del destinatario */
        unsigned int num;       /* numero del mensaje */
};

/* mensaje actual */
struct tagind Msg;

/* destinatario del mensaje */
char To[150];
/* remitente del mensaje */
char Remit[150];
/* remitente (extendido) */
char Xremit[150];
/* subject del mensaje */
char Subject[100];
/* fecha del mensaje */
char Fecha[150];
/* tabla origen del mensaje */
char Tabla[40];
/* gestor de correo origen */
char Mailer[40];

/* linea temporal */
char lineatmp[4096];

/* ancho de la linea en el campo de texto */
int ancho_linea=75;

/* usuario */
char Yo[30]="";

/* fichero de inicializacion (antes era un #define) */
char INI_FILE[80]="wailer.ini";
/* fichero de respuestas */
char RESP_FILE[80]="contesta";
/* fichero de mensajes */
char MSG_FILE[80]="mensajes";

/* estadisticas */
struct
{
        int mensajes;
        int personales;
} Stats;

/* flag que activa el uso del historico; si esta a 0, no se
   guarda nada */
int use_histo=1;


/* estructura para la extraccion de trozos de mensajes */
#define MAX_OBJS 15

struct objfmt
{
        char tipo[20];  /* tipo de codificacin */
        char ini[20];   /* cadena inicial */
        char fin[20];   /* cadena final */
        char ext[4];    /* extensin */
} Objs[MAX_OBJS];


/* indice al siguiente objeto a anadir */
int ObjIndex=0;

/* tipo de objeto hallado en el mensaje */
int ObjType=-1;


/************************
        Funciones
************************/


/** funciones de conferencias */

int LoadConf(void)
/* carga en el array Conferencias las conferencias. Devuelve 0 si no
   se pudo carga el fichero de conferencias */
{
        int i;

        if((i=open(CONFE_FILE, O_RDONLY|O_BINARY))!=-1)
        {
                NumConfes=read(i, Conferencias, MAX_CONFS*MAX_NOM_CONF)
                        /MAX_NOM_CONF;

                close(i);

                return(1);
        }
        else
                return(0);
}


int GetConf(char * conferencia)
/* devuelve el numero asociado a la conferencia. Si no esta en la tabla,
   se genera una nueva y se vuelca la tabla a disco. Si hay algn error
   se devuelve -1 */
{
        int n;
        int o;
        char * ptr;

        /* se salta los prefijos */
        for(ptr=conferencia;*ptr;ptr++)
        {
                if(*ptr=='.')
                {
                        if(memcmp("local",ptr-5,5)==0)
                                *ptr='_';
                        else
                                conferencia=ptr+1;
                }
        }

        /* se recorre las conferencias */
        for(n=0;n<NumConfes;n++)
        {
                if(strcmp(conferencia, Conferencias[n])==0)
                        break;
        }

        /* si esta, pues ese es el numero */
        if(n!=NumConfes)
                return ((unsigned char) n);

        /* aqui llega cuando aun no existe: hay que crearla */

        /* si ya no caben mas (cosa improbable) */
        if((NumConfes+1)==MAX_CONFS)
                return(-1);

        /* se copia */
        strcpy(Conferencias[NumConfes], conferencia);
        NumConfes++;

        /* se abre (o se crea) el fichero de conferencias */
        if((o=open(CONFE_FILE,O_CREAT|O_TRUNC|O_BINARY|O_WRONLY,
                S_IREAD|S_IWRITE))==-1)
                return(-1);

        /* se escribe el array */
        write(o, Conferencias, NumConfes*MAX_NOM_CONF);

        close(o);

        return((unsigned char) (NumConfes-1));
}


/** funciones de mensaje **/

int GoToIndOffset(void)
/* posiciona el puntero al fichero indice sobre el offset del
   tag del mensaje actual. Devuelve el handle del fichero */
{
        int i;
        int off;

        /* se abre el temporal, que contiene los offsets reales */
        if((i=open(TMP_FILE, O_RDONLY|O_BINARY))!=-1)
        {
                lseek(i, Mensaje*sizeof(int), 0);
                read(i,&off,sizeof(int));
                close(i);

                /* se abre el indice */
                if(Histo)
                        i=open(HISTO_IND, O_RDWR|O_BINARY);
                else
                        i=open(NUEVOS_IND, O_RDWR|O_BINARY);

                /* se posiciona */
                lseek(i, off*sizeof(struct tagind), 0);
        }

        return(i);
}


FILE * GoToDatOffset(void)
/* posiciona el puntero sobre el primer caracter de la cabecera de
   un mensaje. Devuelve un FILE * */
{
        int i;
        FILE * f;

        /* coge el offset del tag */
        if((i=GoToIndOffset())==-1)
                return(NULL);

        read(i,&Msg,sizeof(Msg));
        close(i);

        /* se abre el fichero de datos */
        if(Histo==1)
                f=fopen(HISTO_DAT, READ_BINARY);
        else
                f=fopen(NUEVOS_DAT, READ_BINARY);

        return(f);
}


void SetMsgFlags(unsigned char modo)
/* cambia los flags al mensaje actual. Permite intercambiar el flag
   de borrado. Siempre activa el flag de leido */
{
        int i;

        /* se marca como borrado (si mask es MASK_BOR) */
        if(modo&MASK_BOR)
                Msg.dat^=MASK_BOR;

        /* ya leido */
        Msg.dat|=MASK_LEI;

        /* coge la posicion del tag en el indice */
        if((i=GoToIndOffset())!=-1)
        {
                /* se escribe el tag del mensaje */
                write(i, &Msg, sizeof(Msg));
                close(i);
        }
}


char * Cpy(char * ptr, char * lin, int * size,
        void (* procstr)(char * cadena))
/* ejecuta procstr con lin, copia lin sobre ptr, decrementa size */
{
        int s;

        /* si queda sitio... */
        if((*size) && ptr)
        {
                /* se ejecuta la funcion de procesamiento */
                if(procstr!=NULL)
                        (* procstr)(lin);

                /* copia sobre hptr */
                if((s=strlen(lin))>(*size))
                        s=*size;

                memcpy(ptr, lin, s);
                ptr+=s;
                (*size)-=s;
        }

        return(ptr);
}


void GetLogName(char * str)
/* coge de la direccion internet solo el logname */
{
        int n,m;
        char tmp[250];

        strcpy(tmp,str);

        for(n=0;tmp[n]!='@' && tmp[n]!='\n' && tmp[n]!='\0';n++);

        if(tmp[n]!='@')
        {
                str[n]='\0';
                return;
        }

        for(m=1,n--;n>0;m++,n--)
        {
                if(tmp[n]==' ' || tmp[n]=='<')
                {
                        n++;
                        m--;
                        break;
                }
        }

        memcpy(str,&tmp[n],m);
        str[m]='\0';
}


int ProcessMsg(char * hptr, int hsize, char * bptr, int bsize,
        void (* procstr)(char * cadena), int quote)
/* carga la cabecera y el cuerpo del mensaje actual.
   Procstr es una funcion que sera llamada cada vez que
   se lea una linea con la linea leida como argumento, pudiendose
   modificar al gusto (p.e. anadir un \r\n, o lo que sea).
   Devuelve 0 si no se cargo nada. */
{
        FILE * f;
        int n;
        int mm;
        int i;
        char filename[30];

        /* no hay objeto en el mensaje */
        ObjType=-1;

        if((f=GoToDatOffset())==NULL)
                return(0);

	/* se vacian las cadenas */
        Subject[0]=0;
        Fecha[0]=0;
        Tabla[0]=0;
        Mailer[0]=0;

        /* se posiciona sobre la cabecera */
        fseek(f, Msg.offset, 0);

        /* lee la cabecera */
        for(;;)
        {
                if(fgets(lineatmp, sizeof(lineatmp)-1, f)==NULL)
                        break;
                lineatmp[strlen(lineatmp)-1]=0;

                /* si es la ultima de la cabecera */
                if(lineatmp[0]==0)
                        break;

                /* se almacenan los tokens importantes */
                if(memcmp(lineatmp, "Date: ", 6)==0)
                        strcpy(Fecha, lineatmp+6);
                else
                if(memcmp(lineatmp, "X-Tabla: ", 9)==0)
                        strcpy(Tabla, lineatmp+9);
                else
                if(memcmp(lineatmp, "X-Newsreader: ", 14)==0)
                        strcpy(Mailer, lineatmp+14);
                else
                if(memcmp(lineatmp, "From: ", 6)==0)
                        strcpy(Remit, lineatmp+6);
                else
                if(memcmp(lineatmp, "Subject: ", 8)==0)
                        strcpy(Subject, lineatmp+8);

                /* se copia */
                hptr=Cpy(hptr, lineatmp, &hsize, procstr);
        }

        if(hptr)
                *hptr='\0';

        GetLogName(Remit);

        /* se posiciona sobre el cuerpo */
        fseek(f, (Msg.offset+Msg.tamcab)-1, 0);

        /* se lee el numero de bytes que nos dice el tag */
        for(mm=0,n=Msg.tambody;n && bsize>0;)
        {
                /* lee la linea */
                if(fgets(lineatmp, sizeof(lineatmp)-1, f)==NULL)
                        break;
                lineatmp[strlen(lineatmp)-1]=0;

                n-=(strlen(lineatmp)+1);

                /* si se esta procesando un objeto... */
                if(mm>0)
                {
                        write(mm,lineatmp,strlen(lineatmp));
                        write(mm,NL_STR,sizeof(NL_STR)-1);

                        if(memcmp(lineatmp,Objs[ObjType].fin,
                                strlen(Objs[ObjType].fin))==0)
                        {
                                close(mm);
                                mm=0;
                        }
                }
                else
                {
                        /* se lee la linea, controlando si existe
			   algun objeto */
                        for(i=0;i<ObjIndex;i++)
                        {
                                if(memcmp(lineatmp,Objs[i].ini,
                                        strlen(Objs[i].ini))==0)
                                        break;
                        }

                        /* si empieza como uno de los objetos... */
                        if(i<ObjIndex)
                        {
                                ObjType=i;

                                sprintf(filename,"wailer.%s", Objs[ObjType].ext);

                                mm=open(filename,
                                        O_CREAT|O_TRUNC|O_WRONLY|O_BINARY,
                                        S_IREAD|S_IWRITE);

                                write(mm,lineatmp,strlen(lineatmp));
                                write(mm,NL_STR,sizeof(NL_STR)-1);

                                continue;
                        }
                        else
                        {
				/* anade el quote */
                                if(quote)
                                        bptr=Cpy(bptr, QUOTE_CHARS, &bsize, NULL);

                                bptr=Cpy(bptr, lineatmp, &bsize, procstr);
                        }
                }
        }

        if(bptr)
                *bptr='\0';

        fclose(f);

        if(mm)
                close(mm);

        return(1);
}


int ReadMsg(char * hptr, int hsize, char * bptr, int bsize,
        void (* procstr)(char * cadena))
/* lee el mensaje usando ProcessMsg */
{
        return(ProcessMsg(hptr,hsize,bptr,bsize,procstr,0));
}


int QuoteMsg(char * bptr, int bsize, void (* procstr)(char * cadena))
/* rellena ptr con el mensaje actual marcado con > para edicion.
   Tambien modifica el subject. procstr es como en ReadMsg */
{
        int n,c,t;
        char tmp[512];
        int ok;

        if(!ProcessMsg(NULL,(int) NULL,bptr,bsize,procstr,1))
                return(0);

        /* se salta los espacios iniciales en el subject */
        for(c=0;Subject[c]==' ';c++);

        /* anade el Re: */
        if(strncmp(&Subject[c],"Re:",3)==0 ||
           strncmp(&Subject[c],"RE:",3)==0)
        {
                c+=4;

                if(Subject[c]=='(')
                {
                        /* salta el ( */
                        c++;

                        for(n=0;;n++,c++)
                        {
                                if(Subject[c]>='0' && Subject[c]<='9')
                                        tmp[n]=Subject[c];
                                else
                                        break;
                        }
                        tmp[n]=0;

                        t=atoi(tmp)+1;

                        /* salta el ) y el ' ' */
                        c+=2;
                }
                else
                        t=2;

                sprintf(tmp,"Re: (%d) %s", t, &Subject[c]);
        }
        else
                sprintf(tmp,"Re: %s",&Subject[c]);

        strcpy(Subject, tmp);

        return(1);
}


void WriteTag(int handle)
/* escribe un tag en el mensaje */
{
        FILE * f;
        long l;
        long l2;
        int c;
        int oc;
        char tag[2048];
        int n;

        /* si no existe, fuera */
        if((f=fopen(TAG_FILE, READ_BINARY))==0)
                return;

        /* coge el tamano maximo del fichero */
        fseek(f,0,SEEK_END);
        l=ftell(f);

        l2=(long) rand()|(((long) rand())<<16);
        l=l2%l;

        fseek(f,l,SEEK_SET);

        /* busca dos enters */
        for(oc=-1;(c=fgetc(f))!=EOF;oc=c)
        {
                if(c=='\r' && oc=='\n')
                        break;
        }

        /* se escribe en la cadena el tag */
        for(n=0,oc=-1;n<sizeof(tag)-1;n++,oc=c)
        {
                if((c=fgetc(f))==EOF)
                        break;

                if(c=='\r' && oc=='\n')
                        break;

                tag[n]=(char) c;
        }
        tag[n++]='\r';
        tag[n++]='\n';
        tag[n]=0;

        fclose(f);

        /* se escribe */
        write(handle, tag, strlen(tag));
}


void WordWrap(char * ptr)
/* inserta un \n en las palabras que se pasen */
{
        int x;
        char * sp;

        while(*ptr)
        {
                /* si es un quote... no se ajustan las palabras */
                if(*ptr=='>')
                {
                        while(*ptr!='\n' && *ptr!=0)
                                ptr++;
                }
                else
                {
                        /* bucle de lineas */
                        for(x=0,sp=0;*ptr!='\n' && *ptr!=0;ptr++,x++)
                        {
                                if(*ptr==' ' || *ptr=='\t')
                                        sp=ptr;

                                if(x>=ancho_linea)
                                {
                                        if(sp)
                                        {
                                                *sp='\n';
                                                ptr=sp+1;
                                                break;
                                        }
                                }
                        }

                        if(*ptr=='\n')
                                ptr++;
                }
        }
}


void EncodeBase64(unsigned char * des, unsigned char * org, int size)
/* codifica en base64 org sobre des */
{
	int n;

	while(size>0)
	{
		n=(*org)>>2;
		*(des++)=Base64[n];		

		if(size>1)
			n=(((*org)<<4)&0x3f)|(((*(org+1))>>4)&0x3f);
		else
                	n=((*org)<<4)&0x3f;

                org++;
		*(des++)=Base64[n];

		size--;

		if(size>1)
			n=(((*org)<<2)&0x3f)|(((*(org+1))>>6)&0x3f);
		else
		if(size==1)
			n=((*org)<<2)&0x3f;
		else
			n=64;

		org++;
		*(des++)=Base64[n];

		size--;

		if(size>=1)
			n=(*org)&0x3f;
		else
			n=64;

		org++;
		*(des++)=Base64[n];

		size--;
	}

	*des='\0';
}	


int WriteMIMEObject(int o, char * objfile, char * objdesc)
/* escribe objfile como un objeto MIME */
{
	int n;
        int astext;
        FILE * f;
	char linetocode[65];
        char * content_type;

	strlwr(objfile);

	if((f=fopen(objfile, READ_BINARY))==NULL)
		return(0);

	/* busca la extensin */
	for(n=strlen(objfile)-1;n>0;n--)
	{
		if(objfile[n]=='.')
			break;
	}

	astext=0;

        /* decide el tipo de content type */
	if(n==0 || strcmp(&objfile[n],".doc")==0 || strcmp(&objfile[n],".txt")==0 ||
		strcmp(&objfile[n],".c")==0 || strcmp(&objfile[n],".pas")==0)
	{
        	content_type="text/plain";
		astext=1;
        }
	else
	if(strcmp(&objfile[n],".htm")==0 || strcmp(&objfile[n],".html")==0)
	{
		content_type="text/html";
		astext=1;
	}
        else
	if(strcmp(&objfile[n],".gif")==0)
		content_type="image/gif";
	else
	if(strcmp(&objfile[n],".jpg")==0 || strcmp(&objfile[n],".jpeg")==0)
		content_type="image/jpeg";        	
        else
	if(strcmp(&objfile[n],".mpg")==0 || strcmp(&objfile[n],".mpeg")==0)
		content_type="video/mpeg";
	else
		content_type="application/octet-stream";

	if(astext)
	{
		sprintf(lineatmp,
		"--%s%s%s%s; charset=us-ascii%s"\
		"%s8bit%s"\
		"%s%s%s%s",
		WailerBoundary,NL_STR,MimeContentType,content_type,NL_STR,
		MimeConTraEnc,NL_STR,MimeConDes,objdesc,NL_STR,NL_STR);

		write(o,lineatmp,strlen(lineatmp));

		/* escribe todo el fichero */
		for(;;)
		{
			if(fgets(lineatmp,sizeof(lineatmp)-1,f)==NULL)
				break;

			write(o,lineatmp,strlen(lineatmp));
		}
	}
	else
	{ 
		sprintf(lineatmp,
		"--%s%s%s%s; name=%s%s"\
		"%sbase64%s"\
		"%s%s%s%s",
		WailerBoundary,NL_STR,MimeContentType,content_type,
		objfile,NL_STR,
		MimeConTraEnc,NL_STR,
		MimeConDes,objdesc,NL_STR,NL_STR);

		write(o,lineatmp,strlen(lineatmp));

		/* escribe todo el fichero */
		for(;;)
		{
			n=fread(linetocode,1,57,f);

			if(n==0)
                        	break;

                        /* codifica en base64 */
			EncodeBase64(lineatmp,linetocode,n);

			/* anade el fin de linea */
			strcat(lineatmp, NL_STR);

			write(o,lineatmp,strlen(lineatmp));
		}
	}

	/* escribe el boundary final */
	sprintf(lineatmp,
		"--%s--%s%s", WailerBoundary,NL_STR,NL_STR);
	write(o,lineatmp,strlen(lineatmp));

	return(1);
}


int WriteMsg(char * bptr, char * conf, char * objfile, char * objdesc)
/* Envia el mensaje. Los datos se toman de las cadenas globales.
   El contenido de bptr se destruye si existe wailer.fir.
   Devuelve 0 si hubo error */
{
        int i,o;
        time_t t;
        struct tm * ti;

        /* abre el fichero */
        if((o=open(RESP_FILE,
                O_CREAT|O_BINARY|O_WRONLY,S_IREAD|S_IWRITE))==-1)
                return(0);

        /* va al final */
        lseek(o,0,SEEK_END);

        /* conf se recibe porque puede no ser la conferencia actual */
        sprintf(lineatmp,"%sCONFERENCIA: %s%s",NL_STR,conf,NL_STR);
        write(o,lineatmp,strlen(lineatmp));

        sprintf(lineatmp,"Tema: %s%s",Subject, NL_STR);
        write(o,lineatmp,strlen(lineatmp));

        sprintf(lineatmp,"A: %s%s",To,NL_STR);
        write(o,lineatmp,strlen(lineatmp));

        /* se escribe el Message-ID */
        time(&t);
        ti=localtime(&t);
        sprintf(lineatmp,
                "Message-ID: %s_wailer_%s@crcok_%02d%02d%02d_%02d%02d%02d%s",
                Yo, conf, ti->tm_sec, ti->tm_min,
                ti->tm_hour, ti->tm_mday, ti->tm_mon+1, ti->tm_year, NL_STR);
        write(o,lineatmp,strlen(lineatmp));

        /* identificador de cabecera MIME */
	sprintf(lineatmp, "X-MIME-Version: 1.0 (generated by The Wailer)%s",NL_STR);
	write(o,lineatmp,strlen(lineatmp));

	/* si hay objfile, es multipart/mixed; si no, es text/plain */
	if(objfile==NULL)
	{
        	sprintf(lineatmp,
		"%stext/plain; charset=us-ascii%s"\
		"%s8bit%s",
		MimeContentType, NL_STR, MimeConTraEnc, NL_STR);
	}
	else
	{
		sprintf(lineatmp,
		"%smultipart/mixed; boundary=%s%s",
		MimeContentType, WailerBoundary,NL_STR);
	}

        write(o,lineatmp,strlen(lineatmp));

	/* Version del Newsreader: szApp esta definido fuera */
        sprintf(lineatmp,"X-Newsreader: %s%s", szApp,NL_STR);
        write(o,lineatmp,strlen(lineatmp));

        /* linea de separacion entre cabecera y cuerpo */
	sprintf(lineatmp,"%s",NL_STR);
	write(o,lineatmp,strlen(lineatmp));

        /* se cortan las palabras */
        WordWrap(bptr);

	/* si hay objeto, se inserta la cabeza del mensaje */
	if(objfile!=NULL)
	{
		sprintf(lineatmp,
		"--%s%s%s: text/plain; charset=us-ascii%s"\
		"%s8bit%s%s",
		WailerBoundary,NL_STR,MimeContentType,NL_STR,
		MimeConTraEnc,NL_STR,NL_STR);

		write(o,lineatmp,strlen(lineatmp));
	}

        /* se escribe el mensaje */
        write(o,bptr,strlen(bptr));

        /* si hay fichero de firma se anade */
        if((i=open(FIRM_FILE,O_RDONLY|O_BINARY))!=-1)
        {
                write(o,bptr,read(i,bptr,32000));
                close(i);
        }

        sprintf(lineatmp,"%s%s-- %s%s", NL_STR, NL_STR,szApp, NL_STR);
        write(o, lineatmp, strlen(lineatmp));

        WriteTag(o);

        /* anade el fichero pegado */
	if(objfile!=NULL)
		WriteMIMEObject(o, objfile, objdesc);

        /* lo cierra */
        close(o);

        return(1);
}


/** funciones de lista de mensajes **/

int ReadList(void (* newelem)(void))
/* Procesa la lista de mensajes de la conferencia actual,
   rellenando el fichero temporal de offsets
   y llamando a newelem cada vez que se lea un mensaje. Cuando se
   llame a newelem se habrn rellenado Subject, Remit y To */
{
        int i,d,t;
        struct tagind ti;
        int n;
        int num;

        NumMsgs=num=0;

        /* si no hay conferencia... */
        if(Conferencia==-1)
                return(0);

        /* se abren los ficheros */
        if(Histo==1)
        {
                i=open(HISTO_IND, O_RDONLY|O_BINARY);
                d=open(HISTO_DAT, O_RDONLY|O_BINARY);
        }
        else
        {
                i=open(NUEVOS_IND, O_RDONLY|O_BINARY);
                d=open(NUEVOS_DAT, O_RDONLY|O_BINARY);
        }

        if(i==-1 || d==-1)
        {
                close(i);
                close(d);
                return(0);
        }

        /* se crea el fichero temporal */
        if((t=open(TMP_FILE, O_CREAT|O_TRUNC|O_WRONLY|O_BINARY,
                S_IREAD|S_IWRITE))==-1)
        {
                close(i);
                close(d);
                return(0);
        }

        /* bucle de lectura del indice */
        for(;;)
        {
                /* se lee un tag del indice */
                if(read(i, &ti, sizeof(ti))!=sizeof(ti))
                        break;

                /* si no es de esta conferencia, adios */
                if((ti.dat&MASK_CONF)!=Conferencia)
                {
                        num++;
                        continue;
                }

                /* si esta borrado y es el historico, no sale */
                if((ti.dat&MASK_BOR) && Histo)
                {
                        num++;
                        continue;
                }

                /* si este mensaje vale, se guarda en el temporal */
                write(t,&num, sizeof(int));

                if(ti.offsub!=0xFFFF)
                {
                        /* se lee el subject */
                        lseek(d, (long)(ti.offset+ti.offsub), 0);
                        read(d, Subject, sizeof(Subject)-1);
                }
                else
                        Subject[0]=0;

                /* quita caracteres sobrantes */
                for(n=0;Subject[n]!='\n' && Subject[n]!=0 &&
                        n<sizeof(Subject)-1;n++);
                Subject[n]=0;

                if(ti.offfrom!=0xFFFF)
                {
                        /* se lee el from */
                        lseek(d, (long)(ti.offset+ti.offfrom), 0);
                        read(d, Remit, sizeof(Remit)-1);
                }
                else
                        Remit[0]=0;

                GetLogName(Remit);

                if(ti.offto!=0xFFFF)
                {
                        /* se lee el to */
                        lseek(d, (long)(ti.offset+ti.offto), 0);
                        read(d, To, sizeof(To)-1);
                }
                else
                        To[0]=0;

                GetLogName(To);

                num++;
                NumMsgs++;

                /* si existe una funcion de proceso, se la llama */
                if(newelem)
                        (*newelem)();
        }

        close(i);
        close(d);
        close(t);

        return(1);
}


/** funciones de ficheros **/

int Index(char * dat, char * ind)
/* crea un indice .ind segun el fichero .dat */
{
        FILE * nd;
        struct tagind ti;
        int i;
        char tmp[50];
        int n,m;
        char firstline[100];
        long l;
        char * ptr;

        /* abre los datos y crea el indice */
        if((nd=fopen(dat, READ_BINARY))==NULL)
                return(0);

        if((i=open(ind, O_CREAT|O_TRUNC|O_WRONLY|O_BINARY,
                S_IREAD|S_IWRITE))==-1)
        {
                fclose(nd);
                return(0);
        }

        /* lee una linea (porque al principio hay una linea) */
        fgets(lineatmp, sizeof(lineatmp)-1, nd);
        lineatmp[0]=' ';
        lineatmp[1]=0;

        /* lee la primera linea con informacion, para averiguar
           cual es la primera linea de cada mensaje */
        l=ftell(nd);
        fgets(firstline, sizeof(firstline)-1, nd);
        ptr=strchr(firstline, ':');
        if(ptr)
                *ptr=0;
        fseek(nd, l, 0);

        ti.offset=1;
        ti.tamcab=0;
        ti.tambody=0;

        Stats.mensajes=Stats.personales=0;

        for(;;)
        {
                /* se definen los valores iniciales */
                ti.dat=0;
                ti.offset+=ti.tamcab+ti.tambody;
                ti.tamcab=ti.tambody=0;

                ti.offsub=ti.offfrom=ti.offto=0xFFFF;

                /* lee una cabecera */
                for(;;)
                {
                        if(feof(nd))
                                break;

                        /* si la linea no contiene nada,
                           se acabo la cabecera */
                        if(lineatmp[0]==0 || lineatmp[0]=='\n')
                                break;

                        /* compara tokens */
                        if(memcmp("Mensaje ",lineatmp, 8)==0)
                        {
                                /* se coge el numero de mensaje */
                                for(n=0,m=8;lineatmp[m]!=' ';n++,m++)
                                        tmp[n]=lineatmp[m];

                                tmp[n]=0;

                                ti.num=atoi(tmp);
                        }
                        else
                        if(memcmp(firstline, lineatmp, strlen(firstline))==0 ||
                           memcmp("X-BBS:", lineatmp, 6)==0 ||
                           memcmp("X-MMedia:", lineatmp, 9)==0)
                        {
                                /* la linea es X-BBS: <conferencia> [mensaje] */
                                /* por tanto, se busca el < */
                                for(n=0;lineatmp[n]!='<' && lineatmp[n];n++);

                                if(lineatmp[n])
                                {
                                        n++;

                                        /* se lee la conferencia. Antes se leia
                                           de Newsgroups:, pero daba problemas
                                           en las news */
                                        for(m=0;lineatmp[n]!='>' && lineatmp[n];n++,m++)
                                                tmp[m]=lineatmp[n];
                                        tmp[m]=0;

                                        ti.dat=(unsigned char) GetConf(tmp);
                                }

                                /* busca el numero de mensaje */
                                for(n++;lineatmp[n]!='[' && lineatmp[n];n++);

                                if(lineatmp[n])
                                {
                                        n++;

                                        /* se escribe en tmp el numero */
                                        for(m=0;lineatmp[n]!=']' && lineatmp[n];n++,m++)
                                                tmp[m]=lineatmp[n];
                                        tmp[m]=0;

                                        ti.num=atoi(tmp);
                                }
                                else
                                        ti.num=0;
                        }
                        else
                        if(memcmp("From: ", lineatmp, 6)==0)
                        {
                                /* se define el offset del from */
                                ti.offfrom=ti.tamcab+5;
                        }
                        else
                        if(memcmp("To: ", lineatmp, 4)==0)
                        {
                                /* se define el offset del to */
                                ti.offto=ti.tamcab+3;

                                if(memcmp(Yo,lineatmp+4,strlen(Yo))==0)
                                {
                                        Stats.personales++;
                                }
                        }
                        else
                        if(memcmp("Subject: ", lineatmp, 8)==0)
                        {
                                /* se define el offset del subject */
                                ti.offsub=ti.tamcab+7;
                        }

                        /* se anade longitud a la cabecera */
                        ti.tamcab+=strlen(lineatmp);

                        /* lee una linea */
                        fgets(lineatmp, sizeof(lineatmp)-1, nd);
                }

                if(feof(nd))
                        break;

                ti.tamcab+=strlen(lineatmp);

                for(;;)
                {
                        /* lee una linea */
                        fgets(lineatmp, sizeof(lineatmp)-1, nd);

                        if(feof(nd))
                                break;

                        if(memcmp("Mensaje ",lineatmp, 8)==0 ||
                           memcmp("X-BBS:", lineatmp, 6)==0 ||
                           memcmp("X-MMedia:", lineatmp, 9)==0 ||
                           memcmp(firstline,lineatmp,strlen(firstline))==0)
                                break;

                        ti.tambody+=strlen(lineatmp);
                }

                /* escribe el tag del fichero */
                write(i, &ti, sizeof(ti));

                Stats.mensajes++;

                if(feof(nd))
                        break;
        }

        fclose(nd);
        close(i);

        return(1);
}


int DropMsg(void)
/* pasa al historico los mensajes nuevos que se quieren guardar */
{
        int hd,hi,ni,nd;
        struct tagind ti;
        char * ptr;
        int t,n;

        /* se abre el indice de nuevos */
        if((ni=open(NUEVOS_IND,O_RDONLY|O_BINARY))==-1)
                return(0);

        /* datos de nuevos */
        if((nd=open(NUEVOS_DAT,O_RDONLY|O_BINARY))==-1)
        {
                close(ni);
                return(0);
        }

        /* indice al historico */
        if((hi=open(HISTO_IND,O_CREAT|O_BINARY|O_WRONLY,S_IREAD|S_IWRITE))==-1)
        {
                close(ni);
                close(nd);
                return(0);
        }

        /* datos del historico */
        if((hd=open(HISTO_DAT,O_CREAT|O_BINARY|O_WRONLY,S_IREAD|S_IWRITE))==-1)
        {
                close(ni);
                close(nd);
                close(hi);
                return(0);
        }

        /* se va al final del indice y del historico */
        lseek(hi, 0, SEEK_END);
        lseek(hd, 0, SEEK_END);

        for(;;)
        {
                /* lee un tag del indice */
                if(read(ni, &ti, sizeof(ti))!=sizeof(ti))
                        break;

                /* si no esta borrado */
                if(!(ti.dat&MASK_BOR))
                {
                        /* se posiciona en el mensaje */
                        lseek(nd, ti.offset-1, 0);

                        /* se coge el nuevo offset del historico */
                        ti.offset=tell(hd);

                        /* se escribe el cuerpo del mensaje */
                        for(t=ti.tamcab+ti.tambody;t;t-=n)
                        {
                                n=min(t,sizeof(lineatmp)-1);

                                n=read(nd, lineatmp, n);

                                write(hd, lineatmp, n);
                        }

                        /* se escribe en el indice historico */
                        write(hi, &ti, sizeof(ti));
                }
        }

        close(ni);
        close(nd);
        close(hi);
        close(hd);

        return(1);
}


/** funciones de inicializacion **/

void ProcessNew(void)
/* procesa los nuevos mensajes */
{
        /* copia los no borrados al historico */
        if(use_histo)
                DropMsg();

        /* borra los nuevos */
        unlink(NUEVOS_DAT);

        /* renombra los mensajes a los nuevos */
        rename(MSG_FILE, NUEVOS_DAT);

        /* crea el indice de nuevos */
        Index(NUEVOS_DAT, NUEVOS_IND);
}


int IsRespFile(void)
/* mira si existe un fichero de respuestas */
{
        int i;

        if((i=open(RESP_FILE, O_RDONLY|O_BINARY))==-1)
                return(0);
        else
        {
                close(i);
                return(1);
        }
}


void InitRespFile(void)
/* inicializa el fichero de respuestas */
{
        /* simplemente lo borra */
        unlink(RESP_FILE);
}


void AddObj(char * tipo, char * ini, char * fin, char * ext)
/* aade un objeto a la lista de objetos procesables */
{
        int n;

        if(ObjIndex==MAX_OBJS)
                return;

        for(n=0;n<ObjIndex;n++)
        {
                if(strcmp(tipo, Objs[n].tipo)==0)
                        break;
        }

        strcpy(Objs[n].tipo, tipo);
        strcpy(Objs[n].ini, ini);
        strcpy(Objs[n].fin, fin);
        strcpy(Objs[n].ext, ext);

        if(n==ObjIndex)
                ObjIndex++;
}


int InitW(void)
/* inicializa el sistema. Devuelve:
   1 si no hay un paquete nuevo de mensajes.
   0 si no lo hay. */
{
        int i;

        /* aleatoriza random */
        srand((unsigned int) time(NULL));

        /* aade los objetos por defecto en los mensajes */
        AddObj("uuencode","begin ","end","uu");
        AddObj("btoa","xbtoa5","xbtoa End","b2a");

        /* se cargan las conferencias */
        LoadConf();

        /* si no existe un fichero de mensajes... */
        if((i=open(MSG_FILE,O_RDONLY|O_BINARY))==-1)
                return(0);

        close(i);

        return(1);
}


void EndW(void)
/* finaliza */
{
        /* borra el fichero temporal */
        unlink(TMP_FILE);
}

